/*
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.  
 *
 * * Neither the name of the University of California, Berkeley nor
 *   the names of its contributors may be used to endorse or promote
 *   products derived from this software without specific prior 
 *   written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package blog;

import java.util.*;

/**
 * A number random variable.  It consists of a potential object pattern (POP) 
 * and a tuple of generating objects.  Its value in a given world is the 
 * number of objects generated by that POP and those generating objects.
 */
public class NumberVar extends BasicVar {
    /**
     * Creates a new NumberVar for the given POP and given tuple of 
     * generating objects.
     */
    public NumberVar(POP pop, List args) {
	super(args, pop.getArgTypes());
	this.pop = pop;
    }

    /**
     * Returns the potential object pattern associated with this number 
     * variable.
     */
    public POP pop() {
	return pop;
    }

    /**
     * Returns the dependency model associated with this number variable's 
     * POP.
     */
    public DependencyModel getDepModel() {
	return pop.getDepModel();
    }

    /**
     * Returns the variables that stand for the generating objects in this 
     * number variable's dependency model.
     *
     * @return a List of String objects representing the variables
     */
    public List getArgVars() {
	return pop.getGenObjVars();
    }

    /**
     * Returns the integer type.
     */
    public Type getType() {
	return BuiltInTypes.NATURAL_NUM;
    }

    /**
     * Returns the number of objects generated by this POP on this tuple 
     * of generating objects in the given world.
     */
    public int numGenerated(PartialWorld w) {
	return ((Integer) getValue(w)).intValue();
    }

    /**
     * Two NumberVar objects are equal if they have the same POP and
     * their argument lists are equal (recall that list equality is
     * checked by calling the <code>equals</code> method on each
     * corresponding pair of objects in the two lists).
     */
    public boolean equals(Object obj) {
	if (obj instanceof NumberVar) {
	    NumberVar other = (NumberVar) obj;
	    return ((pop() == other.pop()) && (args().equals(other.args())));
	}
	return false;
    }
		
    public int hashCode() {
	return pop().hashCode() ^ args().hashCode();
    }

    /**
     * Returns a string of the form #Type(f1 = o1, ..., fK = oK) 
     * where f1, ..., fK are the generating functions and o1, ..., oK are 
     * the generating objects.
     */
    public String toString() {
	StringBuffer buf = new StringBuffer();
	buf.append("#");
	buf.append(pop.getType());
		
	List generatingFunctions = pop.getGeneratingFunctions();
	if (!generatingFunctions.isEmpty()) {
	    buf.append("(");
	    for (int i = 0; i < generatingFunctions.size(); ++i) {
		buf.append(generatingFunctions.get(i));
		buf.append(" = ");
		buf.append(args().get(i));
		if (i + 1 < generatingFunctions.size()) {
		    buf.append(", ");
		}
	    }
	    buf.append(")");
	}

	return buf.toString();
    }	

    private POP pop;
}
