/*
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.  
 *
 * * Neither the name of the University of California, Berkeley nor
 *   the names of its contributors may be used to endorse or promote
 *   products derived from this software without specific prior 
 *   written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package blog.distrib;

import blog.*;
import java.util.*;
import common.Util;

/**
 * A Binomial distribution with parameters n (number of trials) and
 * p (probability of success for a given trial).  The probability of k
 * successes P(k)= C(n,k) * p^k * (1-p)^(n-k)
 */

public class Binomial implements CondProbDistrib {
    /**
     * Creates a new Binomial with parameters n and p.
     */
    public Binomial(List params) {
	if (!((params.get(0) instanceof Integer) && 
	      (params.get(1) instanceof Number))) {
	    throw new IllegalArgumentException
		("Binomial CPD expects two numerical arguments "
		 + "{n, p} where n is an Integer and 0 <= p <= 1 is a "
		 + "Double. Got: " + params);
	}
	n = ((Integer)params.get(0)).intValue();
	p = ((Number)params.get(1)).doubleValue();
    }

    /**
     * Returns the probability of integer k under this distribution.
     */
    public double getProb(List args, Object value) {
	if (!(value instanceof Integer)) {
	    throw new IllegalArgumentException
		("Binomial CPD defines a distribution over objects" 
		 + " of class Integer, not " + value.getClass() + ".");
	}
	int k = ((Integer)value).intValue();
	return ((Util.factorial(n) / 
		 (Util.factorial(k) * Util.factorial(n - k)))
		* Math.pow(p, k) * Math.pow((1 - p), (n - k)));
    }
    
    /** 
     * Returns the log of the probability of integer k under the distribution.
     */
    public double getLogProb(List args, Object value) {
	if (!(value instanceof Integer)) {
	    throw new IllegalArgumentException
		("Binomial CPD defines a distribution over objects" 
		 + " of class Integer, not " + value.getClass() + ".");
	}
	return Math.log(getProb(args, value));
    }

    /**
     * Returns an integer sampled according to this distribution.  Takes time
     * proprotional to np + 1.  (Reference: Non-Uniform Random Variate 
     * Generation, Devroye http://cgm.cs.mcgill.ca/~luc/rnbookindex.html) 
     * Second time-waiting algorithm.
     */
    public Object sampleVal(List args, Type childType) {
	double q = -Math.log(1 - p);
	double sum = 0;
	int x = 0;
	double u, e;
	while (sum <= q) {
	    u = Util.random();
	    e = -Math.log(u); //exponential random variate
	    sum += (e / (n - x));
	    x += 1;
	}
	return new Integer(x - 1);
    }
    
    
    private int n;
    private double p;
}
